#include <avr/pgmspace.h>
#include "english.h"

// const uint8_t leds[]  = {9, 8, 7, 6, 5, 4, 3, 2, A5, A4, A3};
const uint8_t leds[]  = {A3, A4, A5, 2, 3, 4, 5, 6, 7, 8, 9};


// 110 leds 25/s
// 1000000 microseconds(us) for 2750 LEDs
// meaning 363.63us
// so let's take
//int pwm_off = 5;
//int pwm_on = 120;

uint8_t pwm_off = 0;
uint8_t pwm_on = 64;

uint8_t live_1 = A0;
uint8_t live_2 = A1;
uint8_t live_3 = A2;

uint8_t max_led_on_time = 24;
uint8_t active_led_amount = 0;
uint8_t btn_left = 13;
uint8_t btn_right = 12;
uint8_t btn_fire = 11;

uint8_t buzzer = 10;

uint8_t x_offset = 3;
uint8_t y_offset = 1;
long enemies_time = 0;
long boss_time = 0;
long fighter_time = 0;
long lives_time = 0;

boolean enemy_direction = true;
boolean play_tones = false;

const uint8_t matrix[10][11][2] = {
  {  {1, 0},  {0, 1},  {0, 2},  {0, 3},  {0, 4},  {0, 5},  {0, 6},  {0, 7},  {0, 8},  {0, 9}, {0, 10} }, // 0
  {  {2, 0},  {2, 1},  {1, 2},  {1, 3},  {1, 4},  {1, 5},  {1, 6},  {1, 7},  {1, 8},  {1, 9}, {1, 10} }, // 1
  {  {3, 0},  {3, 1},  {3, 2},  {2, 3},  {2, 4},  {2, 5},  {2, 6},  {2, 7},  {2, 8},  {2, 9}, {2, 10} }, // 2
  {  {4, 0},  {4, 1},  {4, 2},  {4, 3},  {3, 4},  {3, 5},  {3, 6},  {3, 7},  {3, 8},  {3, 9}, {3, 10} }, // 3
  {  {5, 0},  {5, 1},  {5, 2},  {5, 3},  {5, 4},  {4, 5},  {4, 6},  {4, 7},  {4, 8},  {4, 9}, {4, 10} }, // 4
  {  {6, 0},  {6, 1},  {6, 2},  {6, 3},  {6, 4},  {6, 5},  {5, 6},  {5, 7},  {5, 8},  {5, 9}, {5, 10} }, // 5
  {  {7, 0},  {7, 1},  {7, 2},  {7, 3},  {7, 4},  {7, 5},  {7, 6},  {6, 7},  {6, 8},  {6, 9}, {6, 10} }, // 6
  {  {8, 0},  {8, 1},  {8, 2},  {8, 3},  {8, 4},  {8, 5},  {8, 6},  {8, 7},  {7, 8},  {7, 9}, {7, 10} }, // 7
  {  {9, 0},  {9, 1},  {9, 2},  {9, 3},  {9, 4},  {9, 5},  {9, 6},  {9, 7},  {9, 8},  {8, 9}, {8, 10} }, // 8
  { {10, 0}, {10, 1}, {10, 2}, {10, 3}, {10, 4}, {10, 5}, {10, 6}, {10, 7}, {10, 8}, {10, 9}, {9, 10} } // 9
};

uint16_t frame [10] = 
{
  0b00000000000,
  0b01100110110,
  0b01100110110,
  0b01100110000,
  0b01111110110,
  0b01111110110,
  0b01100110110,
  0b01100110110,
  0b01100110110,
  0b00000000000,
};

uint16_t hi [10] = 
{
  0b00000000000,
  0b01100110110,
  0b01100110110,
  0b01100110000,
  0b01111110110,
  0b01111110110,
  0b01100110110,
  0b01100110110,
  0b01100110110,
  0b00000000000,
};

uint16_t enemies[3] =
{ 
  0b11111111,
  0b11111111,
  0b11111111
};

uint16_t shields = 0b00110001100;

uint8_t boss_x = 0;
uint8_t fighter_x = 5;
int8_t lives = 3;

void setup()
{
  Serial.begin(115200);

  pinMode(btn_right, INPUT_PULLUP);
  pinMode(btn_left, INPUT_PULLUP);
  pinMode(btn_fire, INPUT_PULLUP);

  pinMode(live_1, OUTPUT);
  pinMode(live_2, OUTPUT);
  pinMode(live_3, OUTPUT);

  pinMode(buzzer, OUTPUT);

  if (digitalRead(btn_fire) == LOW)
  {
    play_tones = true;
  }
  // testDevice();
}

void delete_frame()
{
  for (uint8_t i=0; i<10; i++)
  {
    frame[i] = 0;
  }
}

void animate_enemies()
{
  if (enemy_direction)
  {
    if (x_offset > 0)
    {
      x_offset--;
    }
    else if (y_offset < 6)
    {
      y_offset++;
      enemy_direction = false;
    }
  }
  else
  {
    if (x_offset < 3)
    {
      x_offset++;
    }
    else if (y_offset < 6)
    {
      y_offset++;
      enemy_direction = true;
    }
  }
  if (y_offset == 6) y_offset = 1;
}

void draw_enemies()
{
   // draw in buffer 

  if (millis() - enemies_time > 1000)
  {
    enemies_time = millis();
    animate_enemies();
    if (play_tones) tone(10, (x_offset) * 110 + 220, 100);
  }
  
  for (uint8_t i = 0; i < 3; i++)
  {
    frame[y_offset + i] = enemies[i] << x_offset;
  }
}

void draw_boss()
{
  if (millis() - boss_time > 100)
  {
    boss_time = millis();
    if (boss_x < 10)
    {
      boss_x++;
    }
    else
    {
      boss_x = 0;
    }
  }
  frame[0] = 1 << boss_x;
}

void draw_shields()
{
  frame[8] = shields;
}

void draw_fighter()
{
  frame[9] = 1 << fighter_x;
}

void draw_lives()
{
  if (millis() - lives_time > 333)
  {
    lives_time = millis();
    lives--;

    if (lives<0) lives = 3;
  }

  if (lives > 2) digitalWrite(live_3, HIGH);
  if (lives > 1) digitalWrite(live_2, HIGH);
  if (lives > 0) digitalWrite(live_1, HIGH);
  delayMicroseconds(30);

  digitalWrite(live_1, LOW);
  digitalWrite(live_2, LOW);
  digitalWrite(live_3, LOW);
}

void loop()
{
  if (millis() < 3000)
  {
    showImage();
  }
  else
  {
    game_routine();
  }
}

void game_routine()
{
  checkButtons();
  delete_frame();
  draw_enemies();
  draw_boss();
  draw_shields();
  draw_lives();
  draw_fighter();
  showImage();
}

void showImage()
{
  active_led_amount = 0;
  for (int y = 0; y < 10; y++)
  {
    for (int x = 0; x < 11; x++)
    {
      if ((frame[y] >> x) & 1)
      {
        charlie_matrix(10 - x, y);
        active_led_amount++;
      }
    }
  }

  for (int z=active_led_amount; z<max_led_on_time; z++)
  {
    delayMicroseconds((pwm_on + pwm_off) / 2);
  }
}

void testDevice()
{
  for (uint8_t y = 0; y < 10; y++)
  {
    for (uint8_t x = 0; x < 11; x++)
    {
      if (play_tones) tone(10, y * 100 + x * 10);
      charlie_matrix_2(x, y, 5);
    }
  }
}

void checkButtons()
{
  if (millis() - fighter_time > 100)
  {
    fighter_time = millis();
    
    if (digitalRead(btn_left) == LOW) {
      if (fighter_x<10) fighter_x++;
    }
  
    if (digitalRead(btn_right) == LOW)
    {
      if (fighter_x>0) fighter_x--;
    }
  }

  if (digitalRead(btn_fire) == LOW)
  {
    
  }
}

void charlie_matrix (int x, int y)
{
  uint8_t cx = matrix[y][x][0];
  uint8_t cy = matrix[y][x][1];

  digitalWrite(leds[cx], LOW);
  pinMode(leds[cx], OUTPUT);
  digitalWrite(leds[cy], HIGH);
  pinMode(leds[cy], OUTPUT);
  delayMicroseconds(pwm_on);

  pinMode(leds[cx], INPUT);
  digitalWrite(leds[cx], LOW);
  pinMode(leds[cy], INPUT);
  digitalWrite(leds[cy], LOW);

  delayMicroseconds(pwm_off);
}

void charlie_matrix_2(uint8_t x, uint8_t y, uint16_t d)
{
  uint8_t cx = matrix[y][x][0];
  uint8_t cy = matrix[y][x][1];

  Serial.print("x: ");
  Serial.print(x);
  Serial.print(", ");
  Serial.print(cx);
  Serial.print(", ");
  Serial.print(leds[cx]);

  Serial.print(" - y: ");
  Serial.print(y);
  Serial.print(", ");
  Serial.print(cy);
  Serial.print(", ");
  Serial.print(leds[cy]);

  Serial.println();

  digitalWrite(leds[cx], LOW);
  pinMode(leds[cx], OUTPUT);
  digitalWrite(leds[cy], HIGH);
  pinMode(leds[cy], OUTPUT);
  delay(d);

  pinMode(leds[cx], INPUT);
  digitalWrite(leds[cx], LOW);
  pinMode(leds[cy], INPUT);
  digitalWrite(leds[cy], LOW);

}
